<?php
/** 
* ezpx installer model
*
* This is the model for ezpx installer script. It gathers current database table/column
* information and checks it against the existing xml file. If necessary, the installer
* runs DDL queries through this model class.
* 
* @package      ezpx 
* @author       Peter Zimon 
*/ 
class Installer extends Model {

  var $necessary_tables;
  var $prefix;
  var $ep_db_xml;
  var $db_model;
  var $actual_db_element;
  var $parser;
  
  
  /**
  * Constructor
  */
  function Installer()
  {
    parent::Model();
    $this->prefix = $this->config->item('db_table_prefix');
    $this->ep_db_xml = './system/application/config/ep_db.xml';
  }
  
  
  /**
  * Parse the ezpx database XML to array
  *
  * @access private
  */
  function _parse_db_xml()
  {
    if (file_exists($this->ep_db_xml))
    {
      $this->parser = xml_parser_create('UTF-8');
      xml_parser_set_option($this->parser, XML_OPTION_CASE_FOLDING, false);
      xml_parser_set_option($this->parser, XML_OPTION_SKIP_WHITE, true);
      xml_set_object($this->parser, $this);
      xml_set_element_handler($this->parser, '_sax_start', '_sax_end');
      xml_set_character_data_handler($this->parser, '_sax_cdata');
      xml_parse($this->parser, file_get_contents($this->ep_db_xml), true);
    }
    else
    {
      $this->db_model = FALSE;
    }
  }
  
  
  /**
  * Opening tag helper function for parser
  * 
  * @param  sax
  * @param  string
  * @param  array
  * @access private
  */
  function _sax_start($sax, $tag, $attr) {
    $this->actual_db_element['cdata'] = '';
    $this->actual_db_element['key'] = $tag;
    switch ($tag)
    {
      case 'table':
        $this->db_model[$this->prefix . $attr['name']] = array();
        $this->actual_db_element['tablename'] = $this->prefix . $attr['name'];
        break;
      case 'field':
        $this->db_model[$this->actual_db_element['tablename']][$attr['name']] = array();
        $this->actual_db_element['fieldname'] = $attr['name'];
        break;
      case 'type':
        $this->db_model[$this->actual_db_element['tablename']][$this->actual_db_element['fieldname']]['type'] = '';
        $this->actual_db_element['cdata'] = 'type';
        break;
      case 'null':
        $this->db_model[$this->actual_db_element['tablename']][$this->actual_db_element['fieldname']]['null'] = '';
        $this->actual_db_element['cdata'] = 'null';
        break;
      case 'key':
        $this->db_model[$this->actual_db_element['tablename']][$this->actual_db_element['fieldname']]['key'] = '';
        $this->actual_db_element['cdata'] = 'key';
        break;
      case 'default':
        $this->db_model[$this->actual_db_element['tablename']][$this->actual_db_element['fieldname']]['default'] = '';
        $this->actual_db_element['cdata'] = 'default';
        break;
      case 'extra':
        $this->db_model[$this->actual_db_element['tablename']][$this->actual_db_element['fieldname']]['extra'] = '';
        $this->actual_db_element['cdata'] = 'extra';
        break;
    }
  }
  
  
  /**
  * Closing tag helper function for parser
  * 
  * @param  sax
  * @param  string
  * @access private
  */
  function _sax_end($sax, $tag) {
    switch ($tag)
    {
      case 'tables':
        break;
    }
  }
  
  
  /**
  * Character data helper function for parser
  * 
  * @param  sax
  * @param  string
  * @access private
  */
  function _sax_cdata($sax, $data) {
    switch ($this->actual_db_element['cdata'])
    {
      case 'type':
        $this->db_model[$this->actual_db_element['tablename']][$this->actual_db_element['fieldname']]['type'] .= trim($data);
        break;
      case 'null':
        $this->db_model[$this->actual_db_element['tablename']][$this->actual_db_element['fieldname']]['null'] .= trim($data);
        break;
      case 'key':
        $this->db_model[$this->actual_db_element['tablename']][$this->actual_db_element['fieldname']]['key'] .= trim($data);
        break;
      case 'default':
        $this->db_model[$this->actual_db_element['tablename']][$this->actual_db_element['fieldname']]['default'] .= trim($data);
        break;
      case 'extra':
        $this->db_model[$this->actual_db_element['tablename']][$this->actual_db_element['fieldname']]['extra'] .= trim($data);
        break;
    }
  }
  
  /**
  * Get current DB model (tables and fields)
  *
  * @access private
  * @return array
  */
  function _get_current_db_model()
  {
    // List tables currently in database
    $tables = $this->db->list_tables();
    $tables = array_flip($tables);
    
    $db_tables = '';
    
    foreach ($tables as $table => $id)
    {
      $query = $this->db->query("DESCRIBE " . $table);
      $result = $query->result_array();
      foreach ($result as $key => $val)
      {
        $db_tables[$table][$val['Field']] = array(
          'type' => $val['Type'],
          'null' => $val['Null'],
          'key' => $val['Key'],
          'default' => $val['Default'],
          'extra' => $val['Extra'],
        );
      }
    }
    
    return $db_tables;
  }
  
  
  /**
  * Check existing tables for installation script
  *
  * @access public
  * @return boolean
  */
  function check_install()
  {
    // Check if photo container folder exists
    if ( ! is_dir('./content/photos'))
    {
      return FALSE;
    }
    
    // Get version DB model
    $this->_parse_db_xml();
    if ($this->db_model == FALSE) return TRUE;
    
    // Get current DB model
    $db_tables = $this->_get_current_db_model();
    if ( ! $db_tables) return FALSE;
    
    // Sort arrays
    ksort($db_tables);
    ksort($this->db_model);
    
    // Compare arrays
    if ($db_tables != $this->db_model)
    {
      $diff = $this->calculate_upgrade_diff();
      
      if (count($diff['tables_to_create']) == 0 && 
          count($diff['fields_to_create']) == 0 && 
          count($diff['fields_to_alter']) == 0 && 
          count($diff['fields_to_drop']) == 0)
      {
        return TRUE;
      }
      
      return FALSE;
    }
    else
    {
      return TRUE;
    }
  }
  
  
  /**
  * Check if clean install is needed
  *
  * @access public
  * @return boolean
  */
  function is_clean_install()
  {
    // Get version DB model
    if ( ! is_array($this->db_model))
    {
      $this->_parse_db_xml();
      if ($this->db_model == FALSE) return TRUE;
    }
    
    // Get current DB model
    $db_tables = $this->_get_current_db_model();
    if ( ! $db_tables) return TRUE;
    
    foreach ($this->db_model as $table => $fields)
    {
      if (is_array($db_tables[$table]))
      {
        return FALSE;
      }
    }
    return TRUE;
  }
  
  
  /**
  * Calculates the differences for the upgrade process
  *
  * @access public
  * @return array
  */
  function calculate_upgrade_diff()
  {
    // Get version DB model
    if ( ! is_array($this->db_model))
    {
      $this->_parse_db_xml();
      if ($this->db_model == FALSE) return FALSE;
    }
    
    // Get current DB model
    $db_tables = $this->_get_current_db_model();
    if ( ! $db_tables) return FALSE;
    
    // Calculate DB tables to be created
    $diff['tables_to_create'] = array();
    $diff['fields_to_create'] = array();
    $diff['fields_to_alter'] = array();
    $diff['fields_to_drop'] = array();
    foreach ($this->db_model as $table => $fields)
    {
      if ( ! isset($db_tables[$table]) && 
           ! in_array($table, $diff['tables_to_create']))
      {
        $diff['tables_to_create'][] = $table;
      }
    }
    
    // Calculate DB fields to be altered or created
    foreach ($this->db_model as $table => $fields)
    {
      if ( ! in_array($table, $diff['tables_to_create']))
      {
        foreach ($fields as $field => $attr)
        {
          if ( ! isset($db_tables[$table][$field]))
          {
            $diff['fields_to_create'][$table][] = $field;
          }
          else
          {
            if (strtolower($db_tables[$table][$field]['type']) != strtolower($attr['type']) OR 
                // strtolower($db_tables[$table][$field]['null']) != strtolower($attr['null']) OR 
                strtolower($db_tables[$table][$field]['key']) != strtolower($attr['key']) OR 
                // strtolower($db_tables[$table][$field]['default']) != strtolower($attr['default']) OR 
                strtolower($db_tables[$table][$field]['extra']) != strtolower($attr['extra']))
            {
              $diff['fields_to_alter'][$table][] = $field;
            }
          }
        }
      }
    }
    
    // Calculate DB fields to be dropped    
    foreach ($db_tables as $table => $fields)
    {
      foreach ($fields as $field => $attr)
      {
        if ( ! isset($this->db_model[$table][$field]) && 
            isset($this->db_model[$table]))
        {
          $diff['fields_to_drop'][$table][] = $field;
        }
      }
    }
    
    return $diff;
  }
  
  
  /**
  * Upgrade ezpx
  *
  * @access public
  * @return boolean
  */
  function do_upgrade()
  {
    // Create photo container folder
    if ( ! is_dir('./content/photos'))
    {
      if ( ! mkdir('./content/photos'))
      {
        return FALSE;
      }
    }
    
    // Create uploaded files container folder
    if ( ! is_dir('./content/uploaded'))
    {
      if ( ! mkdir('./content/uploaded'))
      {
        return FALSE;
      }
    }
    
    $diff = $this->calculate_upgrade_diff();
    
    // Create missing tables
    foreach ($diff['tables_to_create'] as $table)
    {
      if (isset($this->db_model[$table]))
      {
        $fields = $this->db_model[$table];
      }
      
      if ( ! $sql = $this->_lego_create_table($table, $fields))
      {
        echo 'ERROR';
        // return FALSE
      }
      else
      {
        if ( ! $this->db->query($sql))
        {
          echo 'ERROR';
          //return FALSE;
        }
      }
    }
    
    // Create missing fields
    foreach ($diff['fields_to_create'] as $table => $fields)
    {
      if ( ! $sqls = $this->_lego_alter_fields($table, $fields, 'ADD', TRUE))
      {
        return FALSE;
      }
      else
      {
        // EXECUTE SQL's
        foreach ($sqls as $sql)
        {
          $this->db->query($sql);
        }
      }
    }
    
    // Modify fields
    foreach ($diff['fields_to_alter'] as $table => $fields)
    {
      if ( ! $sqls = $this->_lego_alter_fields($table, $fields, 'MODIFY', FALSE))
      {
        return FALSE;
      }
      else
      {
        // EXECUTE SQL's
        foreach ($sqls as $sql)
        {
          $this->db->query($sql);
        }
      }
    }
    
    // Drop fields
    foreach ($diff['fields_to_drop'] as $table => $fields)
    {
      foreach ($fields as $field)
      {
        $sql = 'ALTER TABLE `' . $table . '` DROP `' . $field . '`';
        $this->db->query($sql);
      }
    }
    
    return TRUE;    
  }
  
  
  /**
  * Install ezpx from scratch
  *
  * @access public
  * @return boolean
  */
  function do_clean_install()
  {
    // Create photo container folder
    if ( ! is_dir('./content/photos'))
    {
      if ( ! mkdir('./content/photos'))
      {
        return FALSE;
      }
    }
    
    // Create uploaded files container folder
    if ( ! is_dir('./content/uploaded'))
    {
      if ( ! mkdir('./content/uploaded'))
      {
        return FALSE;
      }
    }
    
    // Get version DB model
    if ( ! is_array($this->db_model))
    {
      $this->_parse_db_xml();
      if ($this->db_model == FALSE) return FALSE;
    }
    
    foreach ($this->db_model as $table => $fields)
    {
      if ( ! $sql = $this->_lego_create_table($table, $fields))
      {
        return FALSE;
      }
      else
      {
        if ( ! $this->db->query($sql))
        {
          return FALSE;
        }
      }
    }    
    return TRUE;
  }
  
  
  /**
  * Lego create or modify table fields SQL query
  *
  * @param  string
  * @param  array
  * @param  string
  * @param  boolean
  * @access private
  * @return array
  */
  function _lego_alter_fields($table, $fields, $procedure, $set_pri_key)
  {
    // Pre-check
    if ( ! isset($table) OR 
         ! isset($fields))
    {
      return FALSE;
    }
    
    if ( ! is_array($fields))
    {
      return FALSE;
    }
    
    $cnt = 0;
    foreach ($fields as $field)
    {
      $sqls[$cnt] = 'ALTER TABLE `' . $table . '` ' . $procedure . ' `' . $field . '`';
      
      // Column definition
      $field_attrs = $this->db_model[$table][$field];
      $def = $this->_lego_column_definition($field_attrs, $set_pri_key);
      $sqls[$cnt] .= $def;
      
      // Set column position
      if ( ! $prev_field = $this->_get_prev_field($table, $field))
      {
        $sqls[$cnt] .= ' FIRST';
      }
      else
      {
        $sqls[$cnt] .= ' AFTER `' . $prev_field . '`';
      }
      
      $cnt++;
      
      // Adding indexes and unique keys
      switch ($field_attrs['key'])
      {
        case 'UNI':
          $sqls[$cnt] = 'ALTER TABLE `' . $table . '` ADD UNIQUE (`' . $field .'`)';
          $cnt++;
          break;
        case 'MUL':
          $sqls[$cnt] = 'ALTER TABLE `' . $table . '` ADD INDEX (`' . $field .'`)';
          $cnt++;
          break;
      }
      
    }
    
    return $sqls;
  }
  
  
  /**
  * Return field's previous field
  *
  * @param  string
  * @param  string
  * @access private
  * @return string
  */
  function _get_prev_field($table, $field)
  {
    $prev_field = '';
    foreach ($this->db_model[$table] as $act_field => $attr)
    {
      if ($act_field == $field)
      {
        break;
      }
      $prev_field = $act_field;
    }
    return $prev_field;
  }
  
  
  /**
  * Lego column definition for a db table column
  * Parameter set_keys is a boolean param indicating whether the primary
  * key should be applied for the field.
  *
  * @param  array
  * @param  boolean
  * @access private
  * @return string
  */
  function _lego_column_definition($attr, $set_key = FALSE)
  {
    // Example: 
    // ALTER TABLE `tmp_1` ADD `id` INT( 11 ) UNSIGNED NOT NULL AUTO_INCREMENT PRIMARY KEY;
    
    $def = " " . $attr['type'];
    
    // Null
    $def .= ($attr['null'] == 'NO') ? " NOT NULL" : " NULL";
    
    // Default
    if (is_numeric($attr['default']) OR
        $attr['default'] == 'NULL')
    {
      $def .= " DEFAULT " . $attr['default'];
    }
    elseif ($attr['default'])
    {
      $def .= " DEFAULT '" . $attr['default'] . "'";
    }
    
    // Extras
    $def .= ($attr['extra']) ? " " . $attr['extra'] : "";
    
    // Key
    if ($set_key)
    {
      if ($attr['key'] == 'PRI')
      {
        $def .= ' PRIMARY KEY';
      }
    }
    
    return $def;
  }
  
  
  /**
  * Lego create table SQL query
  *
  * @param  string
  * @param  array
  * @access private
  * @return string
  */
  function _lego_create_table($table, $fields)
  {
    // Pre-check
    if ( ! isset($table) OR 
         ! isset($fields))
    {
      return FALSE;
    }
    
    if ( ! is_array($fields))
    {
      return FALSE;
    }
    
    // Drop table if exists
    $sql = "DROP TABLE IF EXISTS `" . $table . "`";
    if ( ! $this->db->query($sql))
    {
      return FALSE;
    }
    
    // Lego create table query
    $sql = "CREATE TABLE `" . $table . "` ( \n";
    
    // Get keys
    $key = "";
    $pri = FALSE;
    
    // Primary keys
    foreach ($fields as $field => $attr)
    {
      if ($attr['key'] == 'PRI')
      {
        $key .= "\t";
        if ($pri == TRUE)
        {
          $key .= ", `" . $field . "` ";
        }
        else
        {
          $key .= "PRIMARY KEY (`" . $field . "` ";
        }
        $pri = TRUE;
      }
    }
    if ($pri == TRUE)
    {
      $key .= ")";
    }
    
    // Unique and index keys
    foreach ($fields as $field => $attr)
    {
      if ($key != "" && $attr['key'] != 'PRI' && $attr['key']) 
      {
        $key .= ", \n";
      }
      switch ($attr['key'])
      {
        case 'MUL':
          $key .= "\t";
          $key .= "KEY `" . $field . "` (`" . $field . "`)";
          break;
        case 'UNI':
          $key .= "\t";
          $key .= "UNIQUE KEY `" . $field . "` (`" . $field . "`)";
          break;
      }
    }
    
    reset($fields);
    
    $allfields = count($fields);
    $i = 0;
    
    foreach ($fields as $field => $attr)
    {
      $i++;
      $sql .= "\t";
      
      // Fieldname
      $sql .= "`" . $field . "`";
      
      // Column definition
      $def = $this->_lego_column_definition($attr, FALSE);
      $sql .= $def;
      
      // Ending
      if ($i < $allfields OR 
          $key != "")
      {
        $sql .= ", ";
      }
      $sql .= "\n";
    }
    
    if ($key != "")
    {
      $sql .= $key;
    }
    
    $sql .= ") ENGINE=MyISAM DEFAULT CHARSET=" . $this->db->char_set;
    
    if ($this->db->dbcollat)
    {
      $sql .= " COLLATE=" . $this->db->dbcollat;
    }
    
    return $sql;
  }
  
}
?>